/* Author: Neko <^.^> */
/* If u find any bugs, please contact me on nekochan.3.1415926@gmail.com */

#ifdef __GNUG__
#pragma implementation
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
//#include <sys/time.h>
#include "filters.h"

#include "image3d.h"

namespace i3d {

/*
   Approximation of single precision exp() function. Required for test purposes only.
   Solves glibc bug: https://bugzilla.redhat.com/show_bug.cgi?id=521190
*/
static inline float __Exp(float Val) {
/*
   Val *= 1.442695040f;
   float Clipp = (Val < -126.0) ? -126.0f : Val;
   float Neg = Clipp - ((int)Clipp) + ((float)(Val < 0));
   union {
      uint32_t Int;
      float Float;
   } Result = {(uint32_t)(8388608.0f * (Clipp + 121.2740575f + 27.7280233f / (4.84252568f - Neg) - 1.49012907f * Neg))};
   return Result.Float;
*/
   return std::exp(Val);
}

static inline double __Exp(double Val) {
   return std::exp(Val);
}

/* RGB -> scalar value (Grayscale) conversions */
#define RGB2GRAY(Pixel) \
   (Pixel.red * 0.2989 + 0.5870 * Pixel.green + 0.1141 * Pixel.blue)
#define XYZ2GRAY(Pixel) \
   (Pixel.x * 0.2989 + 0.5870 * Pixel.y + 0.1141 * Pixel.z)

template <class VOXEL> double GetWeight(VOXEL Pixel) {
   return Pixel;
}
template <> double GetWeight(RGB Pixel) {
   return RGB2GRAY(Pixel);
}
template <> double GetWeight(RGB16 Pixel) {
   return RGB2GRAY(Pixel);
}
template <> double GetWeight(Vector3d<float> Pixel) {
   return XYZ2GRAY(Pixel);
}
template <> double GetWeight(Vector3d<double> Pixel) {
   return XYZ2GRAY(Pixel);
}

template <class VOXEL, typename PRECISION> PRECISION GetWeightedDifference(VOXEL Pixel1, VOXEL Pixel2) {
   return std::abs(Pixel1 - Pixel2);
}
template <> float GetWeightedDifference(RGB Pixel1, RGB Pixel2) {
   return static_cast<float> (std::abs(RGB2GRAY(Pixel1) - RGB2GRAY(Pixel2)));
}
template <> float GetWeightedDifference(RGB16 Pixel1, RGB16 Pixel2) {
   return static_cast<float> (std::abs(RGB2GRAY(Pixel1) - RGB2GRAY(Pixel2)));
}
template <> float GetWeightedDifference(Vector3d<float> Pixel1, Vector3d<float> Pixel2) {
   return static_cast<float> (std::abs(XYZ2GRAY(Pixel1) - XYZ2GRAY(Pixel2)));
}
template <> float GetWeightedDifference(Vector3d<double> Pixel1, Vector3d<double> Pixel2) {
   return static_cast<float> (std::abs(XYZ2GRAY(Pixel1) - XYZ2GRAY(Pixel2)));
}
template <> double GetWeightedDifference(RGB Pixel1, RGB Pixel2) {
   return std::abs(RGB2GRAY(Pixel1) - RGB2GRAY(Pixel2));
}
template <> double GetWeightedDifference(RGB16 Pixel1, RGB16 Pixel2) {
   return std::abs(RGB2GRAY(Pixel1) - RGB2GRAY(Pixel2));
}
template <> double GetWeightedDifference(Vector3d<float> Pixel1, Vector3d<float> Pixel2) {
   return std::abs(XYZ2GRAY(Pixel1) - XYZ2GRAY(Pixel2));
}
template <> double GetWeightedDifference(Vector3d<double> Pixel1, Vector3d<double> Pixel2) {
   return std::abs(XYZ2GRAY(Pixel1) - XYZ2GRAY(Pixel2));
}

/* Conditional floating point types for buffers. Depends on image pixel format and selected precision */
template <class ELEMENT, typename PRECISION> struct FPVector{typedef PRECISION TYPE;};
template <> struct FPVector<RGB, float>{typedef Vector3d<float> TYPE;};
template <> struct FPVector<RGB16, float>{typedef Vector3d<float> TYPE;};
template <> struct FPVector<RGB, double>{typedef Vector3d<double> TYPE;};
template <> struct FPVector<RGB16, double>{typedef Vector3d<double> TYPE;};
template <> struct FPVector<Vector3d<float>, float>{typedef Vector3d<float> TYPE;};
template <> struct FPVector<Vector3d<double>, float>{typedef Vector3d<float> TYPE;};
template <> struct FPVector<Vector3d<float>, double>{typedef Vector3d<double> TYPE;};
template <> struct FPVector<Vector3d<double>, double>{typedef Vector3d<double> TYPE;};

template <class ELEMENT> struct Scalar{typedef ELEMENT TYPE;};
template <> struct Scalar<RGB>{typedef RGB::ELEMENT_TYPE TYPE;};
template <> struct Scalar<RGB16>{typedef RGB16::ELEMENT_TYPE TYPE;};
template <> struct Scalar<Vector3d<float> >{typedef float TYPE;};
template <> struct Scalar<Vector3d<double> >{typedef double TYPE;};

/* Assignment functions RGB <-> Vector3d */
template <class R_TYPE, class L_TYPE> inline R_TYPE& __Assign(R_TYPE& R, L_TYPE L) {
   R = (R_TYPE)L;
   return R;
}
template <> inline RGB& __Assign(RGB &R, Vector3d<float> L) {
   R.red = (RGB::ELEMENT_TYPE)(L.x + 0.5f);
   R.green = (RGB::ELEMENT_TYPE)(L.y + 0.5f);
   R.blue = (RGB::ELEMENT_TYPE)(L.z + 0.5f);
   return R;
}
template <> inline RGB16& __Assign(RGB16 &R, Vector3d<float> L) {
   R.red = (RGB16::ELEMENT_TYPE)(L.x + 0.5f);
   R.green = (RGB16::ELEMENT_TYPE)(L.y + 0.5f);
   R.blue = (RGB16::ELEMENT_TYPE)(L.z + 0.5f);
   return R;
}
template <> inline RGB& __Assign(RGB &R, Vector3d<double> L) {
   R.red = (RGB::ELEMENT_TYPE)(L.x + 0.5);
   R.green = (RGB::ELEMENT_TYPE)(L.y + 0.5);
   R.blue = (RGB::ELEMENT_TYPE)(L.z + 0.5);
   return R;
}
template <> inline RGB16& __Assign(RGB16 &R, Vector3d<double> L) {
   R.red = (RGB16::ELEMENT_TYPE)(L.x + 0.5);
   R.green = (RGB16::ELEMENT_TYPE)(L.y + 0.5);
   R.blue = (RGB16::ELEMENT_TYPE)(L.z + 0.5);
   return R;
}
template <> inline Vector3d<float>& __Assign(Vector3d<float> &R, RGB L) {
   R.x = (float)(L.red);
   R.y = (float)(L.green);
   R.z = (float)(L.blue);
   return R;
}
template <> inline Vector3d<float>& __Assign(Vector3d<float> &R, RGB16 L) {
   R.x = (float)(L.red);
   R.y = (float)(L.green);
   R.z = (float)(L.blue);
   return R;
}
template <> inline Vector3d<double>& __Assign(Vector3d<double> &R, RGB L) {
   R.x = (double)(L.red);
   R.y = (double)(L.green);
   R.z = (double)(L.blue);
   return R;
}
template <> inline Vector3d<double>& __Assign(Vector3d<double> &R, RGB16 L) {
   R.x = (double)(L.red);
   R.y = (double)(L.green);
   R.z = (double)(L.blue);
   return R;
}

/* Get time differences = time interval in miliseconds. Used for experimental time mesurements */
/*
double GetElapsedTimeInMs(timeval *Start, timeval *End) {
   unsigned long int FromSec = Start->tv_sec;
   unsigned long int ToSec = End->tv_sec;
   unsigned long int FromuSec = Start->tv_usec;
   unsigned long int TouSec = End->tv_usec;

   if (TouSec < FromuSec) {
      --ToSec;
      TouSec += 1000000;
   }
   ToSec -= FromSec;
   TouSec -= FromuSec;
   ToSec = ToSec * 1000 + TouSec / 1000;
   TouSec %= 1000;
   return (double)ToSec + ((double)TouSec) / 1000;
}
*/

/*
   Recursive Deriche gaussian filter approximation (2nd order)
   
*/
template <typename ELEMENT, typename PRECISION> void BlurDeriche(ELEMENT *Image, int XDim, int YDim, int ZDim, PRECISION Sigma) {
   typedef typename FPVector<ELEMENT, PRECISION>::TYPE FP_TYPE;
   int SizeXY, SizeXYZ;
   int XPos, YPos, ZPos, Idx;
   FP_TYPE IValue, LastIValue, OValue, LastOValue, Value;
   FP_TYPE *Buffer1Pointer, *Buffer2Pointer, *Buffer3Pointer;
   PRECISION Alpha, Alpha0, Alpha1, Alpha2, Alpha3, Beta0, Beta1, ExpAlpha, Coefficient1, Coefficient2;

   SizeXY = XDim * YDim;
   SizeXYZ = SizeXY * ZDim;

   Alpha = ((PRECISION)1.695) / Sigma;
   ExpAlpha = __Exp(-Alpha);
   Beta0 = -((PRECISION)2.0) * ExpAlpha;
   Beta1 = __Exp(-((PRECISION)2) * Alpha);

   Alpha0 = (1 - ExpAlpha) * (1 - ExpAlpha) / (1 - Alpha * Beta0 - Beta1);
   Alpha1 = Alpha0 * (Alpha - 1) * ExpAlpha;
   Alpha2 = Alpha0 * (Alpha + 1) * ExpAlpha;
   Alpha3 = -Alpha0 * Beta1;

   Coefficient1 = (Alpha0 + Alpha1) / (1 + Beta0 + Beta1);
   Coefficient2 = (Alpha2 + Alpha3) / (1 + Beta0 + Beta1);

   Buffer1Pointer = (FP_TYPE *)malloc(SizeXYZ * sizeof(FP_TYPE) * 3);
   if (Buffer1Pointer == NULL) {
      fprintf(stderr, "Out of memory?!\n");
      return;
   }
   Buffer2Pointer = Buffer1Pointer + SizeXYZ;
   Buffer3Pointer = Buffer1Pointer + SizeXYZ * 2;

   for (Idx = 0; Idx < SizeXYZ; ++Idx) {
      __Assign(Buffer1Pointer[Idx], Image[Idx]);
   }
   /* X-axis filter */
   if (XDim > 1) {
      for(ZPos = 0; ZPos < ZDim; ++ZPos) {
         for(YPos = 0; YPos < YDim; ++YPos) {
            Idx = (ZPos * YDim + YPos) * XDim;
            __Assign(LastIValue, Image[Idx]);
            LastOValue = Coefficient1 * LastIValue;
            OValue = LastOValue;
            for (XPos = 0; XPos < XDim; ++XPos) {
               __Assign(IValue, Image[Idx + XPos]);
               Value = Alpha0 * IValue + Alpha1 * LastIValue - Beta0 * OValue - Beta1 * LastOValue;
               Buffer2Pointer[Idx + XPos] = Value;
               LastIValue = IValue;
               LastOValue = OValue;
               OValue = Value;
            }
            __Assign(LastIValue, Image[Idx + XDim - 1]);
            IValue = LastIValue;
            OValue = Coefficient2 * IValue;
            LastOValue = OValue;
            /* Antiextensive pass */
            for (XPos = XDim - 1; XPos >= 0; --XPos) {
               Value = Alpha2 * IValue + Alpha3 * LastIValue - Beta0 * OValue - Beta1 * LastOValue;
               LastIValue = IValue;
               LastOValue = OValue;
               OValue = Value;
               __Assign(IValue, Image[Idx + XPos]);
               Buffer3Pointer[Idx + XPos] = Buffer2Pointer[Idx + XPos] + Value;
            }
         }
      }
   }
   /* Y-axis filtering */
   if (YDim > 1) {
      for(ZPos = 0; ZPos < ZDim; ++ZPos) {
         for(XPos = 0; XPos < XDim; ++XPos) {
            Idx = ZPos * YDim * XDim + XPos;
            LastIValue = Buffer3Pointer[Idx];
            LastOValue = Coefficient1 * LastIValue;
            OValue = LastOValue;
            for (YPos = 0; YPos < SizeXY; YPos += XDim) {
               IValue = Buffer3Pointer[Idx + YPos];
               Value = Alpha0 * IValue + Alpha1 * LastIValue - Beta0 * OValue - Beta1 * LastOValue;
               Buffer2Pointer[Idx + YPos] = Value;
               LastIValue = IValue;
               LastOValue = OValue;
               OValue = Value;
            }
            LastIValue = Buffer3Pointer[Idx + SizeXY - XDim];
            IValue = LastIValue;
            OValue = Coefficient2 * IValue;
            LastOValue = OValue;
            /* Antiextensive pass */
            for (YPos = SizeXY - XDim; YPos >= 0; YPos -= XDim) {
               Value = Alpha2 * IValue + Alpha3 * LastIValue - Beta0 * OValue - Beta1 * LastOValue;
               LastIValue = IValue;
               LastOValue = OValue;
               OValue = Value;
               IValue = Buffer3Pointer[Idx + YPos];
               Buffer1Pointer[Idx + YPos] = Buffer2Pointer[Idx + YPos] + Value;
            }
         }
      }
   } else {
      for (Idx = 0; Idx < SizeXYZ; ++Idx) {
         __Assign(Image[Idx], Buffer3Pointer[Idx]);
      }
   }
   /* Z-axis filtering */
   if (ZDim > 1) {
      for(YPos = 0; YPos < YDim; ++YPos) {
         for(XPos = 0; XPos < XDim; ++XPos) {
            Idx = YPos * XDim + XPos;
            LastIValue = Buffer1Pointer[Idx];
            LastOValue = Coefficient1 * LastIValue;
            OValue = LastOValue;
            for (ZPos = 0; ZPos < SizeXYZ; ZPos += SizeXY) {
               IValue = Buffer1Pointer[Idx + ZPos];
               Value = Alpha0 * IValue + Alpha1 * LastIValue - Beta0 * OValue - Beta1 * LastOValue;
               Buffer2Pointer[Idx + ZPos] = Value;
               LastIValue = IValue;
               LastOValue = OValue;
               OValue = Value;
            }
            LastIValue = Buffer1Pointer[Idx + SizeXYZ - SizeXY];
            IValue = LastIValue;
            OValue = Coefficient2 * IValue;
            LastOValue = OValue;
            /* Antiextensive pass */
            for (ZPos = SizeXYZ - SizeXY; ZPos >= 0; ZPos -= SizeXY) {
               Value = Alpha2 * IValue + Alpha3 * LastIValue - Beta0 * OValue - Beta1 * LastOValue;
               LastIValue = IValue;
               LastOValue = OValue;
               OValue = Value;
               IValue = Buffer1Pointer[Idx + ZPos];
               Buffer3Pointer[Idx + ZPos] = Buffer2Pointer[Idx + ZPos] + Value;
            }
         }
      }
      for (Idx = 0; Idx < SizeXYZ; ++Idx) {
         __Assign(Image[Idx], Buffer3Pointer[Idx]);
      }
   } else {
      for (Idx = 0; Idx < SizeXYZ; ++Idx) {
         __Assign(Image[Idx], Buffer1Pointer[Idx]);
      }
   }

   free(Buffer1Pointer);
}

template <typename ELEMENT> void BlurDeriche_Simple(ELEMENT *Image, int XDim, int YDim, int ZDim, float Sigma) {
   BlurDeriche<ELEMENT, float>(Image, XDim, YDim, ZDim, Sigma);
}
template <typename ELEMENT> void BlurDeriche_Double(ELEMENT *Image, int XDim, int YDim, int ZDim, double Sigma) {
   BlurDeriche<ELEMENT, double>(Image, XDim, YDim, ZDim, Sigma);
}

/* Intensity slices - PBFIC method */
template <class VOXEL, typename PRECISION> void BilateralPBFIC(Image3d<VOXEL>&Image, PRECISION SSigma, PRECISION RSigma, int PBFICCount) {
   typedef typename FPVector<VOXEL, PRECISION>::TYPE FP_TYPE;
   int XDim, YDim, ZDim, Size;
   int XPos, YPos, ZPos, Idx, PBFIC;
   PRECISION Coefficient, RDivisor, Range, Max, Min;
   FP_TYPE Value, *PBFIC_Value, *PBFIC1_Value, *PBFIC2_Value, *TmpPBFICPointer;
   PRECISION *PBFIC_Weight, *PBFIC1_Weight, *PBFIC2_Weight, *TmpWeightPointer;
   PRECISION Intensity, IntensityStep, Divisor;
   VOXEL *ImagePointer, *Buffer;

   XDim = static_cast<int>(Image.GetSizeX());
   YDim = static_cast<int>(Image.GetSizeY());
   ZDim = static_cast<int>(Image.GetSizeZ());
   Size = XDim * YDim * ZDim;

   ImagePointer = Image.begin();
   Min = Max = static_cast<PRECISION> (GetWeight(ImagePointer[0]));
   for (Idx = 1; Idx < Size; ++Idx) {
      Intensity = static_cast<PRECISION> (GetWeight(ImagePointer[Idx]));
      if (Max < Intensity) {
         Max =  Intensity;
      }
      if (Min > Intensity) {
         Min =  Intensity;
      }
   }
   Range = Max - Min;

#ifdef I3D_VERBOSE
   //timeval StartTime, EndTime;
   printf("<<< Required memory: %lu\n", Image.GetImageSize() * (sizeof(VOXEL) * 2 + sizeof(FP_TYPE) * 3 + sizeof(PRECISION) * 2));
#endif
   Buffer = (VOXEL *)malloc(Image.GetImageSize() * sizeof(VOXEL));
   PBFIC_Value = (FP_TYPE *)malloc(Image.GetImageSize() * sizeof(FP_TYPE) * 2);
   PBFIC_Weight = (PRECISION *)malloc(Image.GetImageSize() * sizeof(PRECISION) * 2);
   if (Buffer == NULL || PBFIC_Value == NULL || PBFIC_Weight == NULL) {
      fprintf(stderr, "Out of memory?!\n");
      return;
   }
   PBFIC1_Value = PBFIC_Value;
   PBFIC1_Weight = PBFIC_Weight;
   PBFIC2_Value = PBFIC_Value + Image.GetImageSize();
   PBFIC2_Weight = PBFIC_Weight + Image.GetImageSize();

   RDivisor = ((PRECISION)1.0) / (((PRECISION)2.0) * RSigma * RSigma);
#ifdef I3D_VERBOSE
   //gettimeofday(&StartTime, NULL);
#endif
   /* Create initial slice */
   for (ZPos = 0; ZPos < ZDim; ++ZPos) {
      for (YPos = 0; YPos < YDim; ++YPos) {
         for (XPos = 0; XPos < XDim; ++XPos) {
            Idx = (ZPos * YDim + YPos) * XDim + XPos;
            __Assign(Value, ImagePointer[Idx]);
            Coefficient = (PRECISION)GetWeight<FP_TYPE>(Value) - Min;
            Coefficient = __Exp(-(Coefficient * Coefficient * RDivisor));
            PBFIC2_Value[Idx] = Value * Coefficient;
            PBFIC2_Weight[Idx] = Coefficient;
         }
      }
   }
   /* slice blurring */
   BlurDeriche<FP_TYPE, PRECISION>(PBFIC2_Value, XDim, YDim, ZDim, SSigma);
   BlurDeriche<PRECISION, PRECISION>(PBFIC2_Weight, XDim, YDim, ZDim, SSigma);
   IntensityStep = Range / ((PRECISION)PBFICCount);
   Divisor = ((PRECISION)1.0) / (IntensityStep);
   for (PBFIC = 0, Intensity = IntensityStep + Min; PBFIC < PBFICCount; ++PBFIC, Intensity += IntensityStep) {
      TmpPBFICPointer = PBFIC2_Value;
      PBFIC2_Value = PBFIC1_Value;
      PBFIC1_Value = TmpPBFICPointer;

      TmpWeightPointer = PBFIC2_Weight;
      PBFIC2_Weight = PBFIC1_Weight;
      PBFIC1_Weight = TmpWeightPointer;

      /* Create slice */
      for (ZPos = 0; ZPos < ZDim; ++ZPos) {
         for (YPos = 0; YPos < YDim; ++YPos) {
            for (XPos = 0; XPos < XDim; ++XPos) {
               Idx = (ZPos * YDim + YPos) * XDim + XPos;
               __Assign(Value, ImagePointer[Idx]);
               Coefficient = (PRECISION)GetWeight<FP_TYPE>(Value) - Intensity;
               Coefficient = __Exp(-(Coefficient * Coefficient * RDivisor));
               PBFIC2_Value[Idx] = Value * Coefficient;
               PBFIC2_Weight[Idx] = Coefficient;
            }
         }
      }
      /* slice blurring */
      BlurDeriche<FP_TYPE, PRECISION>(PBFIC2_Value, XDim, YDim, ZDim, SSigma);
      BlurDeriche<PRECISION, PRECISION>(PBFIC2_Weight, XDim, YDim, ZDim, SSigma);
      /* result reconstruction */
      for (ZPos = 0; ZPos < ZDim; ++ZPos) {
         for (YPos = 0; YPos < YDim; ++YPos) {
            for (XPos = 0; XPos < XDim; ++XPos) {
               Idx = (ZPos * YDim + YPos) * XDim + XPos;
               __Assign(Value, ImagePointer[Idx]);
               Coefficient = (PRECISION)GetWeight<FP_TYPE>(Value);
               if ((Coefficient >= Intensity - IntensityStep) && (Coefficient <= Intensity)) {
                  __Assign(Buffer[Idx], ((PBFIC1_Value[Idx] / PBFIC1_Weight[Idx]) * (Intensity - Coefficient) + 
                     (PBFIC2_Value[Idx] / PBFIC2_Weight[Idx]) * (Coefficient - Intensity + IntensityStep)) * Divisor);
               }
            }
         }
      }
   }
   memcpy(ImagePointer, Buffer, Image.GetImageSize() * sizeof(VOXEL));
#ifdef I3D_VERBOSE
   //gettimeofday(&EndTime, NULL);
   //printf("<<< CPU filtering phase only: %.3f ms\n", GetElapsedTimeInMs(&StartTime, &EndTime));
#endif

   /* Clean up */
   free(PBFIC_Value);
   free(PBFIC_Weight);
   free(Buffer);
}

template <class VOXEL> void BilateralPBFIC_Simple(Image3d<VOXEL> &Image, float SSigma, float RSigma, int PBFICCount) {
   BilateralPBFIC<VOXEL, float>(Image, SSigma, RSigma, PBFICCount);
}
template <class VOXEL> void BilateralPBFIC_Double(Image3d<VOXEL> &Image, double SSigma, double RSigma, int PBFICCount) {
   BilateralPBFIC<VOXEL, double>(Image, SSigma, RSigma, PBFICCount);
}

/* Bilateral grid method */
template <class VOXEL, typename PRECISION> void BilateralGrid(Image3d<VOXEL> &Image, PRECISION SSigma, PRECISION RSigma) {
   typedef typename FPVector<VOXEL, PRECISION>::TYPE FP_TYPE;
   typedef typename Scalar<VOXEL>::TYPE S_TYPE;
   const PRECISION GAUSS_5_TAP_CF[] = {static_cast<PRECISION> (10.0), static_cast<PRECISION> (3.6788), static_cast<PRECISION> (0.1)};
   const int GRID_ELEMENT_SIZE = 1 + sizeof(VOXEL) / sizeof(S_TYPE);
   int XDim, YDim, ZDim, Size;
   int XPos, YPos, ZPos, WPos, Idx, Cf, Cnt, SubCnt;
   int GridYDim, GridZDim, GridWDim, GridXDim, GSizeXY, GSizeXYZ;
   PRECISION *Grid;
   PRECISION WValue, Weight, Coefficient, RecRSigma, RecSSigma;
   PRECISION Max, Min, Range;
   PRECISION XOffset, YOffset, ZOffset, ROffset;
   PRECISION GridXPos, GridYPos, GridZPos, GridRPos;
   VOXEL *ImagePointer;
   FP_TYPE Buffer[3], Value;
   PRECISION WBuffer[3];

   XDim = static_cast<int>(Image.GetSizeX());
   YDim = static_cast<int>(Image.GetSizeY());
   ZDim = static_cast<int>(Image.GetSizeZ());
   Size = XDim * YDim * ZDim;

   ImagePointer = Image.begin();
   Min = Max = static_cast<PRECISION> (GetWeight(ImagePointer[0]));
   for (Idx = 1; Idx < Size; ++Idx) {
      WValue = static_cast<PRECISION> (GetWeight(ImagePointer[Idx]));
      if (Max < WValue) {
         Max =  WValue;
      }
      if (Min > WValue) {
         Min =  WValue;
      }
   }
   Range = Max - Min;

   GridYDim = (XDim > 1) ? 1 + (int)ceil(XDim / SSigma) : 1;
   GridZDim = (YDim > 1) ? 1 + (int)ceil(YDim / SSigma) : 1;
   GridWDim = (ZDim > 1) ? 1 + (int)ceil(ZDim / SSigma) : 1;
   GridXDim = 1 + (int)ceil(Range / RSigma);
   GSizeXY = GridXDim * GridYDim;
   GSizeXYZ = GSizeXY * GridZDim;

   XOffset = (GridYDim * SSigma - XDim) * ((PRECISION)0.5);
   YOffset = (GridZDim * SSigma - YDim) * ((PRECISION)0.5);
   ZOffset = (GridWDim * SSigma - ZDim) * ((PRECISION)0.5);
   ROffset = (GridXDim * RSigma - Range) * ((PRECISION)0.5);

   RecRSigma = ((PRECISION)1.0) / RSigma;
   RecSSigma = ((PRECISION)1.0) / SSigma;

#ifdef I3D_VERBOSE
   //timeval StartTime, EndTime;
   printf("<<< Required memory: %lu\n", GridYDim * GridZDim * GridWDim * GridXDim * GRID_ELEMENT_SIZE * sizeof(PRECISION) + Image.GetImageSize() * sizeof(VOXEL));
#endif
   Grid = (PRECISION *)calloc(GridYDim * GridZDim * GridWDim * GridXDim * GRID_ELEMENT_SIZE, sizeof(PRECISION));
   if (Grid == NULL) {
      fprintf(stderr, "Out of memory, increase kernel size or alpha paramete!\n");
      return;
   }
#ifdef I3D_VERBOSE
   //gettimeofday(&StartTime, NULL);
#endif
   //* Grid Creation
   for (ZPos = 0, Cnt = 0; ZPos < ZDim; ++ZPos) {
      for (YPos = 0; YPos < YDim; ++YPos) {
         for (XPos = 0; XPos < XDim; ++XPos, ++Cnt) {
            Idx = (((((int)((ZPos + ZOffset) * RecSSigma)) * GridZDim +
               ((int)((YPos + YOffset) * RecSSigma))) * GridYDim +
               ((int)((XPos + XOffset) * RecSSigma))) * GridXDim +
               ((int)((GetWeight(ImagePointer[Cnt]) - Min + ROffset) * RecRSigma))) * GRID_ELEMENT_SIZE;
            __Assign(Value, ImagePointer[Cnt]);
            *((FP_TYPE *)(Grid + Idx)) += Value;
            ++Grid[Idx + sizeof(VOXEL) / sizeof(S_TYPE)];
         }
      }
   }

   // Gauss along range axis
   if (GridXDim > 1) {
      for (Cnt = 0; Cnt < GSizeXYZ * GridWDim; ++Cnt) {
         Idx = Cnt - (Cnt / GridXDim) * GridXDim;
         Cf = Cnt - (Cnt / 3) * 3;
         Value = (PRECISION)0.0;
         WValue = Weight = (PRECISION)0.0;
         for (XPos = -2; XPos < 3; ++XPos) {
            if (Idx + XPos >= 0 && Idx + XPos < GridXDim) {
               Coefficient = GAUSS_5_TAP_CF[std::abs(XPos)];
               Value += *((FP_TYPE *)(Grid + (Cnt + XPos) * GRID_ELEMENT_SIZE)) * Coefficient;
               WValue += Grid[(Cnt + XPos + 1) * GRID_ELEMENT_SIZE - 1] * Coefficient;
               Weight += Coefficient;
            }
         }
         Buffer[Cf] = Value / Weight;
         WBuffer[Cf] = WValue / Weight;
         if (Idx > 1) {
            *((FP_TYPE *)(Grid + (Cnt - 2) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 1 - ((Cf + 1) / 3) * 3];
            Grid[(Cnt - 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 1 - ((Cf + 1) / 3) * 3];
            if (Idx == GridXDim - 1) {
               *((FP_TYPE *)(Grid + (Cnt - 1) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 2 - ((Cf + 2) / 3) * 3];
               Grid[Cnt * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 2 - ((Cf + 2) / 3) * 3];
               *((FP_TYPE *)(Grid + Cnt * GRID_ELEMENT_SIZE)) = Buffer[Cf];
               Grid[(Cnt + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf];
            }
         }
      }
   }

   // Gauss along X-axis
   if (GridYDim > 1) {
      for (WPos = 0; WPos < GridWDim; ++WPos) {
         for (ZPos = 0; ZPos < GridZDim; ++ZPos) {
            for (XPos = 0; XPos < GridXDim; ++XPos) {
               for (YPos = 0; YPos < GridYDim; ++YPos) {
                  Idx = ((WPos * GridZDim + ZPos) * GridYDim + YPos) * GridXDim + XPos;
                  Cf = YPos - (YPos / 3) * 3;
                  Value = (PRECISION)0.0;
                  WValue = Weight = (PRECISION)0.0;
                  for (Cnt = -2; Cnt < 3; ++Cnt) {
                     if (YPos + Cnt >= 0 && YPos + Cnt < GridYDim) {
                        Coefficient = GAUSS_5_TAP_CF[std::abs(Cf)];
                        Value += *((FP_TYPE *)(Grid + (Idx + Cnt * GridXDim) * GRID_ELEMENT_SIZE)) * Coefficient;
                        WValue += Grid[(Idx + Cnt * GridXDim + 1) * GRID_ELEMENT_SIZE - 1] * Coefficient;
                        Weight += Coefficient;
                     }
                  }
                  Buffer[Cf] = Value / Weight;
                  WBuffer[Cf] = WValue / Weight;
                  if (YPos > 1) {
                     *((FP_TYPE *)(Grid + (Idx - 2 * GridXDim) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                     Grid[(Idx - 2 * GridXDim + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                  }
               }
               *((FP_TYPE *)(Grid + Idx * GRID_ELEMENT_SIZE)) = Buffer[Cf];
               Grid[(Idx + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf];
               if (GridYDim > 2) {
                  *((FP_TYPE *)(Grid + (Idx - GridXDim) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                  Grid[(Idx - GridXDim + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 1 - ((Cf + 1) / 3) * 3];
               } else if (GridYDim == 2) {
                  *((FP_TYPE *)(Grid + (Idx - GridXDim) * GRID_ELEMENT_SIZE)) = Buffer[0];
                  Grid[(Idx - GridXDim + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[0];
               }
            }
         }
      }
   }

   // Gauss along Y-axis
   if (GridZDim > 1) {
      for (WPos = 0; WPos < GridWDim; ++WPos) {
         for (YPos = 0; YPos < GridYDim; ++YPos) {
            for (XPos = 0; XPos < GridXDim; ++XPos) {
               for (ZPos = 0; ZPos < GridZDim; ++ZPos) {
                  Idx = ((WPos * GridZDim + ZPos) * GridYDim + YPos) * GridXDim + XPos;
                  Cf = ZPos - (ZPos / 3) * 3;
                  Value = WValue = Weight = 0.0;
                  for (Cnt = -2; Cnt < 3; ++Cnt) {
                     if (ZPos + Cnt >= 0 && ZPos + Cnt < GridZDim) {
                        Coefficient = GAUSS_5_TAP_CF[std::abs(Cf)];
                        Value += *((FP_TYPE *)(Grid + (Idx + Cnt * GSizeXY) * GRID_ELEMENT_SIZE)) * Coefficient;
                        WValue += Grid[(Idx + Cnt * GSizeXY + 1) * GRID_ELEMENT_SIZE - 1] * Coefficient;
                        Weight += Coefficient;

                     }
                  }
                  Buffer[Cf] = Value / Weight;
                  WBuffer[Cf] = WValue / Weight;
                  if (ZPos > 1) {
                     *((FP_TYPE *)(Grid + (Idx - 2 * GSizeXY) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                     Grid[(Idx - 2 * GSizeXY + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                  }

               }
               *((FP_TYPE *)(Grid + Idx * GRID_ELEMENT_SIZE)) = Buffer[Cf];
               Grid[(Idx + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf];
               if (GridZDim > 2) {
                  *((FP_TYPE *)(Grid + (Idx - GSizeXY) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                  Grid[(Idx - GSizeXY + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 1 - ((Cf + 1) / 3) * 3];
               } else if (GridZDim == 2) {
                  *((FP_TYPE *)(Grid + (Idx - GSizeXY) * GRID_ELEMENT_SIZE)) = Buffer[0];
                  Grid[(Idx - GSizeXY + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[0];
               }
            }
         }
      }
   }

   // Gauss along Z-axis
   if (GridWDim > 1) {
      for (ZPos = 0; ZPos < GridZDim; ++ZPos) {
         for (YPos = 0; YPos < GridYDim; ++YPos) {
            for (XPos = 0; XPos < GridXDim; ++XPos) {
               for (WPos = 0; WPos < GridWDim; ++WPos) {
                  Idx = ((WPos * GridZDim + ZPos) * GridYDim + YPos) * GridXDim + XPos;
                  Cf = WPos - (WPos / 3) * 3;
                  Value = WValue = Weight = 0.0;
                  for (Cnt = -2; Cnt < 3; ++Cnt) {
                     if (WPos + Cnt >= 0 && WPos + Cnt < GridWDim) {
                        Coefficient = GAUSS_5_TAP_CF[std::abs(Cf)];
                        Value += *((FP_TYPE *)(Grid + (Idx + Cnt * GSizeXYZ) * GRID_ELEMENT_SIZE)) * Coefficient;
                        WValue += Grid[(Idx + Cnt * GSizeXYZ + 1) * GRID_ELEMENT_SIZE - 1] * Coefficient;
                        Weight += Coefficient;
                     }
                  }
                  Buffer[Cf] = Value / Weight;
                  WBuffer[Cf] = WValue / Weight;
                  if (WPos > 1) {
                     *((FP_TYPE *)(Grid + (Idx - 2 * GSizeXYZ) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                     Grid[(Idx - 2 * GSizeXYZ + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                  }
               }
               *((FP_TYPE *)(Grid + Idx * GRID_ELEMENT_SIZE)) = Buffer[Cf];
               Grid[(Idx + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf];
               if (GridWDim > 2) {
                  *((FP_TYPE *)(Grid + (Idx - GSizeXYZ) * GRID_ELEMENT_SIZE)) = Buffer[Cf + 1 - ((Cf + 1) / 3) * 3];
                  Grid[(Idx - GSizeXYZ + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[Cf + 1 - ((Cf + 1) / 3) * 3];
               } else if (GridWDim == 2) {
                  *((FP_TYPE *)(Grid + (Idx - GSizeXYZ) * GRID_ELEMENT_SIZE)) = Buffer[0];
                  Grid[(Idx - GSizeXYZ + 1) * GRID_ELEMENT_SIZE - 1] = WBuffer[0];
               }
            }
         }
      }
   }

   // Reconstruction aka slicing
   if (GridWDim > 1) { // 3D data
      for (ZPos = 0, Cf = 0; ZPos < ZDim; ++ZPos) {
         for (YPos = 0; YPos < YDim; ++YPos) {
            for (XPos = 0; XPos < XDim; ++XPos, ++Cf) {
               GridXPos = (XPos + XOffset) * RecSSigma - ((PRECISION)0.5);
               GridYPos = (YPos + YOffset) * RecSSigma - ((PRECISION)0.5);
               GridZPos = (ZPos + ZOffset) * RecSSigma - ((PRECISION)0.5);
               GridRPos = (static_cast<PRECISION> (GetWeight(ImagePointer[Cf])) - Min + ROffset) * RecRSigma - ((PRECISION)0.5);
               Idx = (((int)GridZPos) * GSizeXYZ + ((int)GridYPos) * GSizeXY + ((int)GridXPos) * GridXDim + ((int)GridRPos)) * GRID_ELEMENT_SIZE;
               // Quadlinear interpolation
               GridXPos -= (int)GridXPos;
               GridYPos -= (int)GridYPos;
               GridZPos -= (int)GridZPos;
               GridRPos -= (int)GridRPos;
               Cnt = Idx + GridXDim * GRID_ELEMENT_SIZE;
               WPos = Idx + GSizeXY * GRID_ELEMENT_SIZE;
               SubCnt = WPos + GridXDim * GRID_ELEMENT_SIZE;
               WBuffer[0] = (((Grid[Idx + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[Idx + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((Grid[Cnt + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[Cnt + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * GridXPos)) * (((PRECISION)1.0) - GridYPos);
               Buffer[0] = (((*((FP_TYPE *)(Grid + Idx)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + Idx + GRID_ELEMENT_SIZE)) * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((*((FP_TYPE *)(Grid + Cnt)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + Cnt + GRID_ELEMENT_SIZE)) * GridRPos) * GridXPos)) * (((PRECISION)1.0) - GridYPos);
               WBuffer[1] = (((Grid[WPos + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[WPos + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((Grid[SubCnt + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[SubCnt + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * GridXPos)) * (((PRECISION)1.0) - GridYPos);
               Buffer[1] = (((*((FP_TYPE *)(Grid + WPos)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + WPos + GRID_ELEMENT_SIZE)) * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((*((FP_TYPE *)(Grid + SubCnt)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + SubCnt + GRID_ELEMENT_SIZE)) * GridRPos) * GridXPos)) * (((PRECISION)1.0) - GridYPos);
               Idx += GSizeXYZ * GRID_ELEMENT_SIZE;
               Cnt = Idx + GridXDim * GRID_ELEMENT_SIZE;
               WPos = Idx + GSizeXY * GRID_ELEMENT_SIZE;
               SubCnt = WPos + GridXDim * GRID_ELEMENT_SIZE;
               WBuffer[2] = (((Grid[Idx + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[Idx + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((Grid[Cnt + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[Cnt + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * GridXPos)) * GridYPos + WBuffer[0];
               Buffer[2] = (((*((FP_TYPE *)(Grid + Idx)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + Idx + GRID_ELEMENT_SIZE)) * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((*((FP_TYPE *)(Grid + Cnt)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + Cnt + GRID_ELEMENT_SIZE)) * GridRPos) * GridXPos)) * GridYPos + Buffer[0];
               WValue = (((Grid[WPos + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[WPos + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((Grid[SubCnt + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
                  Grid[SubCnt + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * GridXPos)) * GridYPos + WBuffer[1];
               Value = (((*((FP_TYPE *)(Grid + WPos)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + WPos + GRID_ELEMENT_SIZE)) * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
                  ((*((FP_TYPE *)(Grid + SubCnt)) * (((PRECISION)1.0) - GridRPos) +
                  *((FP_TYPE *)(Grid + SubCnt + GRID_ELEMENT_SIZE)) * GridRPos) * GridXPos)) * GridYPos + Buffer[1];
               Value = (Buffer[2] * (((PRECISION)1.0) - GridZPos) + Value * GridZPos) / 
                  (WBuffer[2] * (((PRECISION)1.0) - GridZPos) + WValue * GridZPos);
               __Assign(ImagePointer[Cf], Value);
            }
         }
      }
   } else if (GridZDim > 1) { // 2D data
      for (YPos = 0, Cf = 0; YPos < YDim; ++YPos) {
         for (XPos = 0; XPos < XDim; ++XPos, ++Cf) {
            GridXPos = (XPos + XOffset) * RecSSigma - ((PRECISION)0.5);
            GridYPos = (YPos + YOffset) * RecSSigma - ((PRECISION)0.5);
            GridRPos = (static_cast<PRECISION> (GetWeight(ImagePointer[Cf])) - Min + ROffset) * RecRSigma - ((PRECISION)0.5);
            Idx = (((int)GridYPos) * GSizeXY + ((int)GridXPos) * GridXDim + ((int)GridRPos)) * GRID_ELEMENT_SIZE;
            // Trilinear interpolation
            GridXPos -= (int)GridXPos;
            GridYPos -= (int)GridYPos;
            GridRPos -= (int)GridRPos;
            Cnt = Idx + GridXDim * GRID_ELEMENT_SIZE;
            WPos = Idx + GSizeXY * GRID_ELEMENT_SIZE;
            SubCnt = WPos + GridXDim * GRID_ELEMENT_SIZE;
            Value = ((((*((FP_TYPE *)(Grid + Idx)) * (((PRECISION)1.0) - GridRPos) +
               *((FP_TYPE *)(Grid + Idx + GRID_ELEMENT_SIZE)) * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
               ((*((FP_TYPE *)(Grid + Cnt)) * (((PRECISION)1.0) - GridRPos) +
               *((FP_TYPE *)(Grid + Cnt + GRID_ELEMENT_SIZE)) * GridRPos) * GridXPos)) * (((PRECISION)1.0) - GridYPos) +
               (((*((FP_TYPE *)(Grid + WPos)) * (((PRECISION)1.0) - GridRPos) +
               *((FP_TYPE *)(Grid + WPos + GRID_ELEMENT_SIZE)) * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
               ((*((FP_TYPE *)(Grid + SubCnt)) * (((PRECISION)1.0) - GridRPos) +
               *((FP_TYPE *)(Grid + SubCnt + GRID_ELEMENT_SIZE)) * GridRPos) * GridXPos)) * GridYPos) /
               ((((Grid[Idx + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
               Grid[Idx + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
               ((Grid[Cnt + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
               Grid[Cnt + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * GridXPos)) * (((PRECISION)1.0) - GridYPos) +
               (((Grid[WPos + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
               Grid[WPos + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
               ((Grid[SubCnt + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
               Grid[SubCnt + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * GridXPos)) * GridYPos);
            __Assign(ImagePointer[Cf], Value);
         }
      }
   } else {
      for (XPos = 0; XPos < XDim; ++XPos) {
         GridXPos = (XPos + XOffset) * RecSSigma - ((PRECISION)0.5);
         GridRPos = (static_cast<PRECISION> (GetWeight(ImagePointer[XPos])) - Min + ROffset) * RecRSigma - ((PRECISION)0.5);
         Idx = (((int)GridXPos) * GridXDim + ((int)GridRPos)) * GRID_ELEMENT_SIZE;
         // Bilinear interpolation
         GridXPos -= (int)GridXPos;
         GridRPos -= (int)GridRPos;
         Cnt = Idx + GridXDim * GRID_ELEMENT_SIZE;
         Value = (((*((FP_TYPE *)(Grid + Idx)) * (((PRECISION)1.0) - GridRPos) +
            *((FP_TYPE *)(Grid + Idx + GRID_ELEMENT_SIZE)) * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
            ((*((FP_TYPE *)(Grid + Cnt)) * (((PRECISION)1.0) - GridRPos) +
            *((FP_TYPE *)(Grid + Cnt + GRID_ELEMENT_SIZE)) * GridRPos) * GridXPos)) /
            (((Grid[Idx + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
            Grid[Idx + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * (((PRECISION)1.0) - GridXPos)) +
            ((Grid[Cnt + GRID_ELEMENT_SIZE - 1] * (((PRECISION)1.0) - GridRPos) +
            Grid[Cnt + GRID_ELEMENT_SIZE * 2 - 1] * GridRPos) * GridXPos));
         __Assign(ImagePointer[XPos], Value);
      }
   }
#ifdef I3D_VERBOSE
   //gettimeofday(&EndTime, NULL);
   //printf("CPU filtering phase only: %.3f ms\n", GetElapsedTimeInMs(&StartTime, &EndTime));
#endif

   // Clean up
   free(Grid);
}

template <class VOXEL> void BilateralGrid_Simple(Image3d<VOXEL> &Image, float SSigma, float RSigma) {
   BilateralGrid<VOXEL, float>(Image, SSigma, RSigma);
}
template <class VOXEL> void BilateralGrid_Double(Image3d<VOXEL> &Image, double SSigma, double RSigma) {
   BilateralGrid<VOXEL, double>(Image, SSigma, RSigma);
}

/* Naive method (terribly slow !!) */
template <class VOXEL, typename PRECISION> void BilateralNaive(Image3d<VOXEL> &Image, PRECISION SSigma, PRECISION RSigma) {
   typedef typename FPVector<VOXEL, PRECISION>::TYPE FP_TYPE;
   int XDim, YDim, ZDim;
   int XPos, YPos, ZPos, CfX, CfY, CfZ, Idx;
   int KernelWidth, KernelHeight, KernelDepth;
   PRECISION Weight, Coefficient;
   PRECISION RDivisor, SDivisor;
   FP_TYPE Sum, RefValue, Value;
   VOXEL *ImagePointer, *Buffer;

   XDim = static_cast<int>(Image.GetSizeX());
   YDim = static_cast<int>(Image.GetSizeY());
   ZDim = static_cast<int>(Image.GetSizeZ());

   KernelDepth = KernelHeight = KernelWidth = (int)(SSigma * 3.0f);
   if (KernelWidth > XDim) {
      KernelWidth = XDim;
   }
   if (KernelHeight > YDim) {
      KernelHeight = YDim;
   }
   if (KernelDepth > ZDim) {
      KernelDepth = ZDim;
   }
#ifdef I3D_VERBOSE
   //timeval StartTime, EndTime;
   printf("<<< Required memory: %lu\n", Image.GetImageSize() * sizeof(VOXEL) * 2);
#endif

   /* Bilateral filtering */
   SDivisor = ((PRECISION)1.0) / (((PRECISION)2.0) * SSigma * SSigma);
   RDivisor = ((PRECISION)1.0) / (((PRECISION)2.0) * RSigma * RSigma);
   ImagePointer = Image.begin();
   Buffer = (VOXEL *)malloc(Image.GetImageSize() * sizeof(VOXEL));
#ifdef I3D_VERBOSE
   //gettimeofday(&StartTime, NULL);
#endif
   for (ZPos = 0; ZPos < ZDim; ++ZPos) {
      for (YPos = 0; YPos < YDim; ++YPos) {
         for (XPos = 0; XPos < XDim; ++XPos) {
            Sum = Weight = (PRECISION)0.0;
            Idx = (ZPos * YDim + YPos) * XDim + XPos;
            __Assign(RefValue, ImagePointer[Idx]);
            for (CfZ = 1 - KernelDepth; CfZ < KernelDepth; ++CfZ) {
               for (CfY = 1 - KernelHeight; (CfY < KernelHeight) && (ZPos + CfZ >= 0 && ZPos + CfZ < ZDim); ++CfY) {
                  for (CfX = 1 - KernelWidth; (CfX < KernelWidth) && (YPos + CfY >= 0 && YPos + CfY < YDim); ++CfX) {
                     if (XPos + CfX >= 0 && XPos + CfX < XDim) {
                        __Assign(Value, ImagePointer[(CfZ * YDim + CfY) * XDim + CfX + Idx]);
                        Coefficient = GetWeightedDifference<FP_TYPE, PRECISION>(Value, RefValue);
                        Coefficient = __Exp(-(CfX * CfX + CfY * CfY + CfZ * CfZ) * SDivisor -(Coefficient * Coefficient) * RDivisor);
                        Sum += Value * Coefficient;
                        Weight += Coefficient;
                     }
                  }
               }
            }
            __Assign(Buffer[Idx], Sum / Weight);
         }
      }
   }
   /* Replace image structure data with filtered one */
   memcpy(ImagePointer, Buffer, Image.GetImageSize() * sizeof(VOXEL));
#ifdef I3D_VERBOSE
   //gettimeofday(&EndTime, NULL);
   //printf("\nCPU filtering phase only: %.3f ms\n", GetElapsedTimeInMs(&StartTime, &EndTime));
#endif
   free(Buffer);
}

template <class VOXEL> void BilateralNaive_Simple(Image3d<VOXEL> &Image, float SSigma, float RSigma) {
   BilateralNaive<VOXEL, float>(Image, SSigma, RSigma);
}
template <class VOXEL> void BilateralNaive_Double(Image3d<VOXEL> &Image, double SSigma, double RSigma) {
   BilateralNaive<VOXEL, double>(Image, SSigma, RSigma);
}

/* Recursive variant */
template <class VOXEL, typename PRECISION> void BilateralRecursive(Image3d<VOXEL> &Image, PRECISION SSigma, PRECISION RSigma) {
   const PRECISION SQRT2 = (PRECISION)1.4142135623730950488016887242097;
   typedef typename FPVector<VOXEL, PRECISION>::TYPE FP_TYPE;
   Image3d<FP_TYPE> Buffer1, Buffer2;
   int XDim, YDim, ZDim;
   int XPos, YPos, ZPos, Idx, SubIdx;
   PRECISION RDivisor, SDivisor, Coefficient;
   FP_TYPE LastValue, Value, *Buffer1Pointer, *Buffer2Pointer;
   VOXEL *ImagePointer;

   XDim = static_cast<int>(Image.GetSizeX());
   YDim = static_cast<int>(Image.GetSizeY());
   ZDim = static_cast<int>(Image.GetSizeZ());

   Buffer1.CopyMetaData(Image);
   Buffer2.CopyMetaData(Image);
   ImagePointer = Image.begin();
   Buffer1Pointer = Buffer1.begin();
   Buffer2Pointer = Buffer2.begin();

#ifdef I3D_VERBOSE
   //timeval StartTime, EndTime;
   printf("<<< Required memory: %lu\n", Image.GetImageSize() * (sizeof(VOXEL) + sizeof(FP_TYPE) * 2));
#endif

   for (int i = 0; i < XDim * YDim * ZDim; ++i) {
      __Assign(Buffer1Pointer[i], ImagePointer[i]);
   }

   RDivisor = ((PRECISION)1.0) / ((PRECISION(1.5)) * RSigma);
   SDivisor = __Exp(-SQRT2 / (((PRECISION)3.0) * SSigma));
#ifdef I3D_VERBOSE
   //gettimeofday(&StartTime, NULL);
#endif
   /* X-axis filter */
   if (XDim > 1) {
      for(ZPos = 0; ZPos < ZDim; ++ZPos) {
         for(YPos = 0; YPos < YDim; ++YPos) {
            Idx = (ZPos * YDim + YPos) * XDim;
            Buffer2Pointer[Idx] = Buffer1Pointer[Idx];
            /* Extensive pass (Left => Right) */
            for(XPos = 1; XPos < XDim; ++XPos) {
               ++Idx;
               Coefficient = __Exp(-GetWeightedDifference<FP_TYPE, PRECISION>(Buffer1Pointer[Idx], Buffer1Pointer[Idx - 1]) * RDivisor) * SDivisor;
               Buffer2Pointer[Idx] = (((PRECISION)1.0) - Coefficient) * Buffer1Pointer[Idx] + Coefficient * Buffer2Pointer[Idx - 1];
            }
            LastValue = Buffer1Pointer[Idx];
            Buffer2Pointer[Idx] = (Buffer2Pointer[Idx] + LastValue) * ((PRECISION)0.5);
            /* Antiextensive pass (Right => Left) */
            for(XPos = XDim - 2; XPos >= 0; --XPos) {
               --Idx;
               Coefficient = __Exp(-GetWeightedDifference<FP_TYPE, PRECISION>(Buffer1Pointer[Idx], Buffer1Pointer[Idx + 1]) * RDivisor) * SDivisor;
               Value = (((PRECISION)1.0) - Coefficient) * Buffer1Pointer[Idx] + Coefficient * LastValue;
               Buffer2Pointer[Idx] = (Buffer2Pointer[Idx] + Value) * ((PRECISION)0.5);
               LastValue = Value;
            }
         }
      }
      /* Y-axis filter */
      if (YDim > 1) {
         for(ZPos = 0; ZPos < ZDim; ++ZPos) {
            /* Extensive pass (Top => Bottom) */
            for (int i = 0; i < XDim; ++i) {
               Buffer1Pointer[ZPos * XDim * YDim + i] = Buffer2Pointer[ZPos * XDim * YDim + i];
            }
            for(YPos = 1; YPos < YDim; ++YPos) {
               for(XPos = 0; XPos < XDim; ++XPos) {
                  Idx = (ZPos * YDim + YPos) * XDim + XPos;
                  SubIdx = Idx - XDim;
                  Coefficient = __Exp(-GetWeightedDifference<FP_TYPE, PRECISION>(Buffer2Pointer[Idx], Buffer2Pointer[SubIdx]) * RDivisor) * SDivisor;
                  Buffer1Pointer[Idx] = (((PRECISION)1.0) - Coefficient) * Buffer2Pointer[Idx] + Coefficient * Buffer1Pointer[SubIdx];
               }
            }
            for(XPos = 0; XPos < XDim; ++XPos) {
               ++SubIdx;
               Buffer1Pointer[SubIdx] = (Buffer1Pointer[SubIdx] + Buffer2Pointer[SubIdx]) * ((PRECISION)0.5);
            }
            /* Antiextensive pass (Bottom => Top) */
            for(YPos = YDim - 2; YPos >= 0; --YPos) {
               for (XPos = 0; XPos < XDim; ++XPos) {
                  Idx = (ZPos * YDim + YPos) * XDim + XPos;
                  SubIdx = Idx + XDim;
                  Coefficient = __Exp(-GetWeightedDifference<FP_TYPE, PRECISION>(Buffer2Pointer[Idx], Buffer2Pointer[SubIdx]) * RDivisor) * SDivisor;
                  Buffer1Pointer[Idx] = (Buffer1Pointer[Idx] + (((PRECISION)1.0) - Coefficient) * Buffer2Pointer[Idx] + Coefficient * Buffer2Pointer[SubIdx]) * ((PRECISION)0.5);
               }
            }
         }
         /* Z-axis filter */
         if (ZDim > 1) {
            /* Extensive pass (Near => Far) */
            for (int i = 0; i < XDim * YDim; ++i) {
               Buffer2Pointer[i] = Buffer1Pointer[i];
            }
            for(ZPos = 1; ZPos < ZDim; ++ZPos) {
               for(YPos = 0; YPos < YDim; ++YPos) {
                  for(XPos = 0; XPos < XDim; ++XPos) {
                     Idx = (ZPos * YDim + YPos) * XDim + XPos;
                     SubIdx = Idx - YDim * XDim;
                     Coefficient = __Exp(-GetWeightedDifference<FP_TYPE, PRECISION>(Buffer1Pointer[Idx], Buffer1Pointer[SubIdx]) * RDivisor) * SDivisor;
                     Buffer2Pointer[Idx] = (((PRECISION)1.0) - Coefficient) * Buffer1Pointer[Idx] + Coefficient * Buffer2Pointer[SubIdx];
                  }
               }
            }
            for(YPos = 0; YPos < YDim; ++YPos) {
               for(XPos = 0; XPos < XDim; ++XPos) {
                  ++SubIdx;
                  Buffer2Pointer[SubIdx] = (Buffer2Pointer[SubIdx] + Buffer1Pointer[SubIdx]) * ((PRECISION)0.5);
               }
            }
            /* Antiextensive pass (Far => Near) */
            for(ZPos = ZDim - 2; ZPos >= 0; --ZPos) {
               for(YPos = 0; YPos < YDim; ++YPos) {
                  for (XPos = 0; XPos < XDim; ++XPos) {
                     Idx = (ZPos * YDim + YPos) * XDim + XPos;
                     SubIdx = Idx + YDim * XDim;
                     Coefficient = __Exp(-GetWeightedDifference<FP_TYPE, PRECISION>(Buffer1Pointer[Idx], Buffer1Pointer[SubIdx]) * RDivisor) * SDivisor;
                     Buffer2Pointer[Idx] = (Buffer2Pointer[Idx] + (((PRECISION)1.0) - Coefficient) * Buffer1Pointer[Idx] + Coefficient * Buffer1Pointer[SubIdx]) * ((PRECISION)0.5);
                  }
               }
            }
            for (int i = 0; i < XDim * YDim * ZDim; ++i) {
               __Assign(ImagePointer[i], Buffer2Pointer[i]);
            }
         } else {
            for (int i = 0; i < XDim * YDim * ZDim; ++i) {
               __Assign(ImagePointer[i], Buffer1Pointer[i]);
            }
         }
      } else {
         for (int i = 0; i < XDim * YDim * ZDim; ++i) {
            __Assign(ImagePointer[i], Buffer2Pointer[i]);
         }
      }
   }
#ifdef I3D_VERBOSE
   //gettimeofday(&EndTime, NULL);
   //printf("CPU filtering phase only: %.3f ms\n", GetElapsedTimeInMs(&StartTime, &EndTime));
#endif
}

template <class VOXEL> void BilateralRecursive_Simple(Image3d<VOXEL> &Image, float SSigma, float RSigma) {
   BilateralRecursive<VOXEL, float>(Image, SSigma, RSigma);
}
template <class VOXEL> void BilateralRecursive_Double(Image3d<VOXEL> &Image, double SSigma, double RSigma) {
   BilateralRecursive<VOXEL, double>(Image, SSigma, RSigma);
}

/* Approximation via separability */
template <class VOXEL, typename PRECISION> void BilateralSeparation(Image3d<VOXEL> &Image, PRECISION SSigma, PRECISION RSigma) {
   typedef typename FPVector<VOXEL, PRECISION>::TYPE FP_TYPE;
   int XDim, YDim, ZDim;
   int XPos, YPos, ZPos, Cf, Idx;
   int KernelWidth, KernelHeight, KernelDepth;
   PRECISION Weight, Coefficient;
   PRECISION RDivisor, SDivisor;
   FP_TYPE Sum, RefValue, Value, *Buffer1, *Buffer2;
   VOXEL *ImagePointer;

   XDim = static_cast<int>(Image.GetSizeX());
   YDim = static_cast<int>(Image.GetSizeY());
   ZDim = static_cast<int>(Image.GetSizeZ());

   KernelDepth = KernelHeight = KernelWidth = (int)(SSigma * 3.0f);
   if (KernelWidth > XDim) {
      KernelWidth = XDim;
   }
   if (KernelHeight > YDim) {
      KernelHeight = YDim;
   }
   if (KernelDepth > ZDim) {
      KernelDepth = ZDim;
   }

#ifdef I3D_VERBOSE
   //timeval StartTime, EndTime;
   printf("<<< Required memory: %lu\n", Image.GetImageSize() * (sizeof(VOXEL) + sizeof(FP_TYPE) * 2));
#endif

   SDivisor = ((PRECISION)1.0) / (((PRECISION)2.0) * SSigma * SSigma);
   RDivisor = ((PRECISION)1.0) / (((PRECISION)2.0) * RSigma * RSigma);

   Buffer1 = (FP_TYPE *)malloc(Image.GetImageSize() * sizeof(FP_TYPE) * 2);
   if (Buffer1 == NULL) {
      fprintf(stderr, "Out of memory?!\n");
      return;
   }
   Buffer2 = Buffer1 + Image.GetImageSize();
   ImagePointer = Image.begin();
#ifdef I3D_VERBOSE
   //gettimeofday(&StartTime, NULL);
#endif
   /* Filter along X-axis */
   if (XDim > 1) {
      for (ZPos = 0; ZPos < ZDim; ++ZPos) {
         for (YPos = 0; YPos < YDim; ++YPos) {
            for (XPos = 0; XPos < XDim; ++XPos) {
               Sum = Weight = (PRECISION(0.0));
               Idx = (ZPos * YDim + YPos) * XDim + XPos;
               __Assign(RefValue, ImagePointer[Idx]);
               for (Cf = -KernelWidth + 1; Cf < KernelWidth; ++Cf) {
                  if (XPos + Cf >= 0 && XPos + Cf < XDim) {
                     __Assign(Value, ImagePointer[Idx]);
                     Coefficient = (PRECISION)GetWeightedDifference<FP_TYPE, PRECISION>(Value, RefValue);
                     Coefficient = __Exp(-Cf * Cf * SDivisor - Coefficient * Coefficient * RDivisor);
                     Sum += Value * Coefficient;
                     Weight += Coefficient;
                  }
               }
               Buffer1[Idx] = Sum / Weight;
            }
         }
      }
      /* Filter along Y-axis */
      if (YDim > 1) {
         for (ZPos = 0; ZPos < ZDim; ++ZPos) {
            for (YPos = 0; YPos < YDim; ++YPos) {
               for (XPos = 0; XPos < XDim; ++XPos) {
                  Sum = Weight = (PRECISION(0.0));
                  Idx = (ZPos * YDim + YPos) * XDim + XPos;
                  RefValue = Buffer1[Idx];
                  for (Cf = -KernelHeight + 1; Cf < KernelHeight; ++Cf) {
                     if (YPos + Cf >= 0 && YPos + Cf < YDim) {
                        Value = Buffer1[Cf * XDim + Idx];
                        Coefficient = (PRECISION)GetWeightedDifference<FP_TYPE, PRECISION>(Value, RefValue);
                        Coefficient = __Exp(-Cf * Cf * SDivisor - Coefficient * Coefficient * RDivisor);
                        Sum += Value * Coefficient;
                        Weight += Coefficient;
                     }
                  }
                  Buffer2[Idx] = Sum / Weight;
               }
            }
         }
         /* Filter along Z-axis */
         if (ZDim > 1) {
            for (ZPos = 0; ZPos < ZDim; ++ZPos) {
               for (YPos = 0; YPos < YDim; ++YPos) {
                  for (XPos = 0; XPos < XDim; ++XPos) {
                     Sum = Weight = (PRECISION(0.0));
                     Idx = (ZPos * YDim + YPos) * XDim + XPos;
                     RefValue = Buffer2[Idx];
                     for (Cf = -KernelDepth + 1; Cf < KernelDepth; ++Cf) {
                        if (ZPos + Cf >= 0 && ZPos + Cf < ZDim) {
                           Value = Buffer2[Cf * YDim * XDim + Idx];
                           Coefficient = (PRECISION)GetWeightedDifference<FP_TYPE, PRECISION>(Value, RefValue);
                           Coefficient = __Exp(-Cf * Cf * SDivisor - Coefficient * Coefficient * RDivisor);
                           Sum += Value * Coefficient;
                           Weight += Coefficient;
                        }
                     }
                     Buffer1[Idx] = Sum / Weight;
                  }
               }
            }
            for (int i = 0; i < XDim * YDim * ZDim; ++i) {
               __Assign(ImagePointer[i], Buffer1[i]);
            }
         } else {
            for (int i = 0; i < XDim * YDim * ZDim; ++i) {
               __Assign(ImagePointer[i], Buffer2[i]);
            }
         }
      } else {
         for (int i = 0; i < XDim * YDim * ZDim; ++i) {
            __Assign(ImagePointer[i], Buffer1[i]);
         }
      }
   }
#ifdef I3D_VERBOSE
   //gettimeofday(&EndTime, NULL);
   //printf("CPU filtering phase only: %.3f ms\n", GetElapsedTimeInMs(&StartTime, &EndTime));
#endif

   /* Clean up */
   free(Buffer1);
}

template <class VOXEL> void BilateralSeparation_Simple(Image3d<VOXEL> &Image, float SSigma, float RSigma) {
   BilateralSeparation<VOXEL, float>(Image, SSigma, RSigma);
}
template <class VOXEL> void BilateralSeparation_Double(Image3d<VOXEL> &Image, double SSigma, double RSigma) {
   BilateralSeparation<VOXEL, double>(Image, SSigma, RSigma);
}

/* Quality testing (MSE/PSNR) */
template <typename ELEMENT> void SetToMin(ELEMENT &Val1, ELEMENT Val2) {
   Val1 = (Val1 < Val2) ? Val1 : Val2;
}
template <> void SetToMin(RGB &Val1, RGB Val2) {
   Val1.red = (Val1.red < Val2.red) ? Val1.red : Val2.red;
   Val1.green = (Val1.green < Val2.green) ? Val1.green : Val2.green;
   Val1.blue = (Val1.blue < Val2.blue) ? Val1.blue : Val2.blue;
}
template <> void SetToMin(RGB16 &Val1, RGB16 Val2) {
   Val1.red = (Val1.red < Val2.red) ? Val1.red : Val2.red;
   Val1.green = (Val1.green < Val2.green) ? Val1.green : Val2.green;
   Val1.blue = (Val1.blue < Val2.blue) ? Val1.blue : Val2.blue;
}
template <> void SetToMin(Vector3d<float> &Val1, Vector3d<float> Val2) {
   Val1.x = (Val1.x < Val2.x) ? Val1.x : Val2.x;
   Val1.y = (Val1.y < Val2.y) ? Val1.y : Val2.y;
   Val1.z = (Val1.z < Val2.z) ? Val1.z : Val2.z;
}
template <> void SetToMin(Vector3d<double> &Val1, Vector3d<double> Val2) {
   Val1.x = (Val1.x < Val2.x) ? Val1.x : Val2.x;
   Val1.y = (Val1.y < Val2.y) ? Val1.y : Val2.y;
   Val1.z = (Val1.z < Val2.z) ? Val1.z : Val2.z;
}

template <typename ELEMENT> void SetToMax(ELEMENT &Val1, ELEMENT Val2) {
   Val1 = (Val1 > Val2) ? Val1 : Val2;
}
template <> void SetToMax(RGB &Val1, RGB Val2) {
   Val1.red = (Val1.red > Val2.red) ? Val1.red : Val2.red;
   Val1.green = (Val1.green > Val2.green) ? Val1.green : Val2.green;
   Val1.blue = (Val1.blue > Val2.blue) ? Val1.blue : Val2.blue;
}
template <> void SetToMax(RGB16 &Val1, RGB16 Val2) {
   Val1.red = (Val1.red > Val2.red) ? Val1.red : Val2.red;
   Val1.green = (Val1.green > Val2.green) ? Val1.green : Val2.green;
   Val1.blue = (Val1.blue > Val2.blue) ? Val1.blue : Val2.blue;
}
template <> void SetToMax(Vector3d<float> &Val1, Vector3d<float> Val2) {
   Val1.x = (Val1.x > Val2.x) ? Val1.x : Val2.x;
   Val1.y = (Val1.y > Val2.y) ? Val1.y : Val2.y;
   Val1.z = (Val1.z > Val2.z) ? Val1.z : Val2.z;
}
template <> void SetToMax(Vector3d<double> &Val1, Vector3d<double> Val2) {
   Val1.x = (Val1.x > Val2.x) ? Val1.x : Val2.x;
   Val1.y = (Val1.y > Val2.y) ? Val1.y : Val2.y;
   Val1.z = (Val1.z > Val2.z) ? Val1.z : Val2.z;
}

template I3D_ALGO_EXPORT void BilateralNaive_Simple<GRAY8>(Image3d<GRAY8>&, float, float);
template I3D_ALGO_EXPORT void BilateralNaive_Double<GRAY8>(Image3d<GRAY8>&, double, double);
template I3D_ALGO_EXPORT void BilateralNaive_Simple<GRAY16>(Image3d<GRAY16>&, float, float);
template I3D_ALGO_EXPORT void BilateralNaive_Double<GRAY16>(Image3d<GRAY16>&, double, double);
template I3D_ALGO_EXPORT void BilateralNaive_Simple<RGB>(Image3d<RGB>&, float, float);
template I3D_ALGO_EXPORT void BilateralNaive_Double<RGB>(Image3d<RGB>&, double, double);
template I3D_ALGO_EXPORT void BilateralNaive_Simple<RGB16>(Image3d<RGB16>&, float, float);
template I3D_ALGO_EXPORT void BilateralNaive_Double<RGB16>(Image3d<RGB16>&, double, double);
template I3D_ALGO_EXPORT void BilateralNaive_Simple<float>(Image3d<float>&, float, float);
template I3D_ALGO_EXPORT void BilateralNaive_Double<double>(Image3d<double>&, double, double);

template I3D_ALGO_EXPORT void BilateralRecursive_Simple<GRAY8>(Image3d<GRAY8>&, float, float);
template I3D_ALGO_EXPORT void BilateralRecursive_Double<GRAY8>(Image3d<GRAY8>&, double, double);
template I3D_ALGO_EXPORT void BilateralRecursive_Simple<GRAY16>(Image3d<GRAY16>&, float, float);
template I3D_ALGO_EXPORT void BilateralRecursive_Double<GRAY16>(Image3d<GRAY16>&, double, double);
template I3D_ALGO_EXPORT void BilateralRecursive_Simple<RGB>(Image3d<RGB>&, float, float);
template I3D_ALGO_EXPORT void BilateralRecursive_Double<RGB>(Image3d<RGB>&, double, double);
template I3D_ALGO_EXPORT void BilateralRecursive_Simple<RGB16>(Image3d<RGB16>&, float, float);
template I3D_ALGO_EXPORT void BilateralRecursive_Double<RGB16>(Image3d<RGB16>&, double, double);
template I3D_ALGO_EXPORT void BilateralRecursive_Simple<float>(Image3d<float>&, float, float);
template I3D_ALGO_EXPORT void BilateralRecursive_Double<double>(Image3d<double>&, double, double);

template I3D_ALGO_EXPORT void BilateralSeparation_Simple<GRAY8>(Image3d<GRAY8>&, float, float);
template I3D_ALGO_EXPORT void BilateralSeparation_Double<GRAY8>(Image3d<GRAY8>&, double, double);
template I3D_ALGO_EXPORT void BilateralSeparation_Simple<GRAY16>(Image3d<GRAY16>&, float, float);
template I3D_ALGO_EXPORT void BilateralSeparation_Double<GRAY16>(Image3d<GRAY16>&, double, double);
template I3D_ALGO_EXPORT void BilateralSeparation_Simple<RGB>(Image3d<RGB>&, float, float);
template I3D_ALGO_EXPORT void BilateralSeparation_Double<RGB>(Image3d<RGB>&, double, double);
template I3D_ALGO_EXPORT void BilateralSeparation_Simple<RGB16>(Image3d<RGB16>&, float, float);
template I3D_ALGO_EXPORT void BilateralSeparation_Double<RGB16>(Image3d<RGB16>&, double, double);
template I3D_ALGO_EXPORT void BilateralSeparation_Simple<float>(Image3d<float>&, float, float);
template I3D_ALGO_EXPORT void BilateralSeparation_Double<double>(Image3d<double>&, double, double);

template I3D_ALGO_EXPORT void BilateralGrid_Simple<GRAY8>(Image3d<GRAY8>&, float, float);
template I3D_ALGO_EXPORT void BilateralGrid_Double<GRAY8>(Image3d<GRAY8>&, double, double);
template I3D_ALGO_EXPORT void BilateralGrid_Simple<GRAY16>(Image3d<GRAY16>&, float, float);
template I3D_ALGO_EXPORT void BilateralGrid_Double<GRAY16>(Image3d<GRAY16>&, double, double);
template I3D_ALGO_EXPORT void BilateralGrid_Simple<RGB>(Image3d<RGB>&, float, float);
template I3D_ALGO_EXPORT void BilateralGrid_Double<RGB>(Image3d<RGB>&, double, double);
template I3D_ALGO_EXPORT void BilateralGrid_Simple<RGB16>(Image3d<RGB16>&, float, float);
template I3D_ALGO_EXPORT void BilateralGrid_Double<RGB16>(Image3d<RGB16>&, double, double);
template I3D_ALGO_EXPORT void BilateralGrid_Simple<float>(Image3d<float>&, float, float);
template I3D_ALGO_EXPORT void BilateralGrid_Double<double>(Image3d<double>&, double, double);

template I3D_ALGO_EXPORT void BilateralPBFIC_Simple<GRAY8>(Image3d<GRAY8>&, float, float, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Double<GRAY8>(Image3d<GRAY8>&, double, double, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Simple<GRAY16>(Image3d<GRAY16>&, float, float, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Double<GRAY16>(Image3d<GRAY16>&, double, double, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Simple<RGB>(Image3d<RGB>&, float, float, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Double<RGB>(Image3d<RGB>&, double, double, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Simple<RGB16>(Image3d<RGB16>&, float, float, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Double<RGB16>(Image3d<RGB16>&, double, double, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Simple<float>(Image3d<float>&, float, float, int);
template I3D_ALGO_EXPORT void BilateralPBFIC_Double<double>(Image3d<double>&, double, double, int);

template I3D_ALGO_EXPORT void BlurDeriche_Simple(GRAY8 *, int, int, int, float);
template I3D_ALGO_EXPORT void BlurDeriche_Double(GRAY8 *, int, int, int, double);
template I3D_ALGO_EXPORT void BlurDeriche_Simple(RGB16 *, int, int, int, float);
template I3D_ALGO_EXPORT void BlurDeriche_Double(RGB16 *, int, int, int, double);
template I3D_ALGO_EXPORT void BlurDeriche_Simple(GRAY16 *, int, int, int, float);
template I3D_ALGO_EXPORT void BlurDeriche_Double(GRAY16 *, int, int, int, double);
template I3D_ALGO_EXPORT void BlurDeriche_Simple(RGB *, int, int, int, float);
template I3D_ALGO_EXPORT void BlurDeriche_Double(RGB *, int, int, int, double);
template I3D_ALGO_EXPORT void BlurDeriche_Simple(size_t *, int, int, int, float);
template I3D_ALGO_EXPORT void BlurDeriche_Double(size_t *, int, int, int, double);
//template I3D_ALGO_EXPORT void BlurDeriche_Simple(unsigned long *, int, int, int, float);
//template I3D_ALGO_EXPORT void BlurDeriche_Double(unsigned long *, int, int, int, double);
template I3D_ALGO_EXPORT void BlurDeriche_Simple(float *, int, int, int, float);
template I3D_ALGO_EXPORT void BlurDeriche_Double(double *, int, int, int, double);

}
