/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * node.cc
 *
 * Heap node
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 */

#include "node.h"

namespace i3d
{
	namespace fls
	{
		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                       Struct Node                           //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		template <class KEY> bool Node<KEY>::Update(KEY value)
		{
			if (m_key != value)
			{
				m_key = value;
				return true;
			}

			return false;
		}

		/**********************************************************************/

		/////////////////////////////////////////////////////////////////
		//                                                             //
		//                       Struct NHA_Node		               //
		//                                                             //
		/////////////////////////////////////////////////////////////////

		/**********************************************************************/

		NHA_Node::NHA_Node(Type::Size index, Type::Prec arrival_time) 
			: Node<Type::Prec>(index, arrival_time), m_arrival_time(arrival_time), m_speed(Max()) 
		{
		}

		/**********************************************************************/
		
		NHA_Node::NHA_Node(Type::Size index, Type::Prec arrival_time, Type::Prec speed) 
			: Node<Type::Prec>(index, arrival_time + PosReciproc(speed)), m_arrival_time(arrival_time), m_speed(speed)
		{
		}

		/**********************************************************************/

		bool NHA_Node::Update(Type::Prec value)
		{
			if (m_speed != value)
			{
				m_speed = value;
				m_key = m_arrival_time + PosReciproc(value);
				return true;
			}

			return false;
		}

		/**********************************************************************/

	} // namespace fls 

} // namespace i3d

